import os
import shutil
import uuid

from biolib.biolib_logging import logger_no_user_data
from biolib.compute_node.job_worker.cache_types import LfsCacheStateDict, UuidStr, StoragePartition, \
    DockerImageCacheStateDict
from biolib.typing_utils import Dict, List
from biolib.utils.cache_state import CacheState, CacheStateError


class DockerCacheStateError(CacheStateError):
    pass


class LfsCacheState(CacheState):

    def __init__(self):
        super().__init__()

        self._storage_paths: List[str] = self._get_storage_paths()
        self._tmp_storage_paths: List[str] = self._get_tmp_storage_paths()

    @property
    def storage_paths(self) -> List[str]:
        return self._storage_paths

    @property
    def tmp_storage_paths(self) -> List[str]:
        return self._tmp_storage_paths

    @property
    def main_lfs_storage_path(self) -> str:
        return self._storage_paths[0]

    @property
    def _state_path(self) -> str:
        state_path = os.environ.get('BIOLIB_LFS_STATE_PATH')
        if not state_path:
            raise CacheStateError('Environment variable "BIOLIB_LFS_STATE_PATH" not set')

        return state_path

    def _get_default_state(self) -> LfsCacheStateDict:
        return LfsCacheStateDict(
            large_file_systems={},
            storage_partitions=self._get_storage_partitions_from_env(),
        )

    @staticmethod
    def _get_tmp_storage_paths() -> List[str]:
        lfs_tmp_storage_path_env_key = 'BIOLIB_LFS_TMP_STORAGE_PATHS'
        lfs_tmp_storage_paths = os.environ.get(lfs_tmp_storage_path_env_key)
        if not lfs_tmp_storage_paths:
            raise CacheStateError(f'Environment variable "{lfs_tmp_storage_path_env_key}" not set')

        lfs_tmp_storage_paths_list = lfs_tmp_storage_paths.split(',')
        for lfs_tmp_storage_path in lfs_tmp_storage_paths_list:
            if not os.path.isdir(lfs_tmp_storage_path):
                raise CacheStateError(f'LFS temporary storage path {lfs_tmp_storage_path} is not a directory')

        return lfs_tmp_storage_paths_list

    @staticmethod
    def _get_storage_paths() -> List[str]:
        lfs_storage_path_env_key = 'BIOLIB_LFS_STORAGE_PATHS'
        lfs_storage_paths = os.environ.get(lfs_storage_path_env_key)
        logger_no_user_data.debug(f'LFS storage paths: {lfs_storage_paths}')

        # It is essential to check like this so we catch if it is None and if it is empty string
        if not lfs_storage_paths:
            raise CacheStateError(f'Environment variable "{lfs_storage_path_env_key}" not set')

        lfs_storage_paths_list = lfs_storage_paths.split(',')
        for lfs_storage_path in lfs_storage_paths_list:
            if not os.path.isdir(lfs_storage_path):
                raise CacheStateError(f'LFS storage path {lfs_storage_path} is not a directory')

        return lfs_storage_paths_list

    def _get_storage_partitions_from_env(self) -> Dict[UuidStr, StoragePartition]:
        storage_states: Dict[UuidStr, StoragePartition] = {}
        for lfs_storage_path in self._storage_paths:
            uuid_str = str(uuid.uuid4())
            disk_usage = shutil.disk_usage(lfs_storage_path)
            storage_states[uuid_str] = StoragePartition(
                allocated_size_bytes=0,
                path=lfs_storage_path,
                total_size_bytes=disk_usage.free,
                uuid=uuid_str,
            )

        return storage_states


class DockerImageCacheState(CacheState):
    @property
    def _state_path(self) -> str:
        return f'{self._user_cache_dir}/docker-cache-state.json'

    def _get_default_state(self) -> DockerImageCacheStateDict:
        return {}



from typing import Optional


class BioLibError(Exception):

    def __init__(self, message):
        super().__init__(message)
        self.message = message


class NotFound(BioLibError):

    def __init__(self, message: Optional[str] = None):
        super().__init__(message=message or 'Not found.')


class DockerContainerNotFoundDuringExecutionException(Exception):
    pass


class RetryLimitException(Exception):
    pass


class StorageDownloadFailed(Exception):
    pass


class CloudJobFinishedError(Exception):
    pass



import os

import requests

from biolib import utils
from biolib.biolib_api_client import BiolibApiClient, CreatedJobDict
from biolib.biolib_api_client.biolib_job_api import BiolibJobApi
from biolib.biolib_binary_format.utils import RemoteIndexableBuffer
from biolib.biolib_binary_format.unencrypted_module_output import UnencryptedModuleOutput
from biolib.compute_node.cloud_utils import CloudUtils
from biolib.biolib_logging import logger_no_user_data
from biolib.utils.multipart_uploader import get_chunk_iterator_from_file_object


class JobStorage:
    module_output_file_name = 'module-output.bbf'
    legacy_module_output_file_name = 'legacy-module-output.bbf'

    @staticmethod
    def upload_module_input(job: CreatedJobDict, module_input_serialized: bytes) -> None:
        base_url = BiolibApiClient.get().base_url
        job_uuid = job['public_id']
        headers = {'Job-Auth-Token': job['auth_token']}

        multipart_uploader = utils.MultiPartUploader(
            start_multipart_upload_request=dict(
                requires_biolib_auth=False,
                url=f'{base_url}/api/jobs/{job_uuid}/storage/input/start_upload/',
                headers=headers
            ),
            get_presigned_upload_url_request=dict(
                requires_biolib_auth=False,
                url=f'{base_url}/api/jobs/{job_uuid}/storage/input/presigned_upload_url/',
                headers=headers
            ),
            complete_upload_request=dict(
                requires_biolib_auth=False,
                url=f'{base_url}/api/jobs/{job_uuid}/storage/input/complete_upload/',
                headers=headers
            ),
        )

        multipart_uploader.upload(
            payload_iterator=utils.get_chunk_iterator_from_bytes(module_input_serialized),
            payload_size_in_bytes=len(module_input_serialized),
        )

    @staticmethod
    def upload_module_output(job_uuid: str, job_temporary_dir: str) -> None:
        logger_no_user_data.debug(f'Job "{job_uuid}" uploading result...')
        module_output_path = os.path.join(job_temporary_dir, JobStorage.module_output_file_name)
        module_output_size = os.path.getsize(module_output_path)

        with open(module_output_path, mode='rb') as module_output_file:
            module_output_iterator = get_chunk_iterator_from_file_object(module_output_file)
            multipart_uploader = JobStorage._get_module_output_uploader(job_uuid)
            multipart_uploader.upload(
                payload_iterator=module_output_iterator,
                payload_size_in_bytes=module_output_size,
            )

        logger_no_user_data.debug(f'Job "{job_uuid}" result uploaded successfully')

    @staticmethod
    def _get_module_output_uploader(job_uuid: str) -> utils.MultiPartUploader:
        base_url = BiolibApiClient.get().base_url
        config = CloudUtils.get_webserver_config()
        compute_node_auth_token = config['compute_node_info']['auth_token']  # pylint: disable=unsubscriptable-object
        headers = {'Compute-Node-Auth-Token': compute_node_auth_token}

        return utils.MultiPartUploader(
            start_multipart_upload_request=dict(
                requires_biolib_auth=False,
                url=f'{base_url}/api/jobs/{job_uuid}/storage/results/start_upload/',
                headers=headers,
            ),
            get_presigned_upload_url_request=dict(
                requires_biolib_auth=False,
                url=f'{base_url}/api/jobs/{job_uuid}/storage/results/presigned_upload_url/',
                headers=headers,
            ),
            complete_upload_request=dict(
                requires_biolib_auth=False,
                url=f'{base_url}/api/jobs/{job_uuid}/storage/results/complete_upload/',
                headers=headers,
            ),
        )

    @staticmethod
    def get_module_input(job: CreatedJobDict) -> bytes:
        job_uuid = job['public_id']
        logger_no_user_data.debug(f'Job "{job_uuid}" downloading module input...')
        presigned_download_url = BiolibJobApi.get_job_storage_download_url(
            job_uuid=job_uuid,
            job_auth_token=job['auth_token'],
            storage_type='input',
        )
        response = requests.get(url=presigned_download_url)
        response.raise_for_status()
        data: bytes = response.content
        logger_no_user_data.debug(f'Job "{job_uuid}" module input downloaded')
        return data

    @staticmethod
    def get_module_output(job: CreatedJobDict) -> UnencryptedModuleOutput:
        presigned_download_url = BiolibJobApi.get_job_storage_download_url(
            job_uuid=job['public_id'],
            job_auth_token=job['auth_token'],
            storage_type='results',
        )

        buffer = RemoteIndexableBuffer(url=presigned_download_url)
        return UnencryptedModuleOutput(buffer)



import os
import shutil
import sys
import tempfile
import time
from pathlib import Path
from typing import Tuple
from zipfile import ZipFile

import yaml

from biolib.biolib_api_client import AppVersionOnJob, Module, FilesMapping
from biolib.validators.validate_app_version import validate_app_version  # type: ignore
from biolib.validators.validate_argument import validate_argument  # type: ignore
from biolib.validators.validate_module import validate_module  # type: ignore
from biolib.biolib_errors import BioLibError
from biolib.biolib_logging import logger
from biolib.typing_utils import Dict, List


class BiolibValidationError(Exception):
    pass


def validate_and_get_args(args):
    if args is None:
        args = []

    if isinstance(args, str):
        args = list(filter(lambda p: p != '', args.split(' ')))

    if not isinstance(args, list):
        raise Exception('The given input arguments must be list or str')

    return args


def get_files_dict_and_file_args(files, args):
    if files is None:
        files = []
        # TODO: Figure out how to make this slightly less error prone
        for idx, arg in enumerate(args):
            if not arg.startswith('/'):
                arg = f'{os.getcwd()}/{arg}'

            arg = Path(arg)
            if arg.is_file():
                files.append(arg)
                # Make sure that arg is only the filename
                args[idx] = arg.parts[-1]

    files_dict = {}
    for file_path in files:
        file = open(file_path, 'rb')
        path = '/' + file_path.parts[-1]

        files_dict[path] = file.read()
        file.close()

    return files_dict, args


def validate_and_get_app_path(provided_app_path):
    if provided_app_path.startswith('/'):
        app_path = Path(provided_app_path)
    else:
        app_path = Path(os.getcwd()) / provided_app_path

    if not app_path.exists():
        print(f'ERROR: Could not find path {app_path}')
        sys.exit(1)

    if not app_path.is_dir():
        print(f'ERROR: Path {app_path} is not a directory')
        sys.exit(1)

    return app_path


def get_yaml_data(app_path):
    try:
        yaml_file = open(f'{app_path}/.biolib/config.yml', 'r', encoding='utf-8')

    except Exception as error:  # pylint: disable=broad-except
        raise BioLibError(f'Could not open the biolib config file at {app_path}/.biolib/config.yml') from error

    try:
        yaml_data = yaml.safe_load(yaml_file)

    except Exception as error:  # pylint: disable=broad-except
        raise BioLibError(f'Could not parse {app_path}/.biolib/config.yml. Please make sure it is valid YAML') \
            from error

    return yaml_data


def validate_arguments(yaml_data):
    error_dict = {'arguments': {}}
    # Arguments are optional
    if 'arguments' in yaml_data:
        for argument in yaml_data['arguments']:
            argument_errors = validate_argument(argument)
            if argument_errors:
                error_dict['arguments'].update(argument_errors)

    # Just return empty dict if we have no errors
    if error_dict['arguments']:
        return error_dict
    else:
        return {}


def validate_and_get_biolib_yaml_version(yaml_data, error_dict):
    if not 'biolib_version' in yaml_data.keys():
        error_dict['config_yml']['biolib_version'] = ['Your config file is missing the biolib_version field.']
        print_errors(error_dict)
        sys.exit(1)
    else:
        biolib_version = yaml_data['biolib_version']

    if biolib_version not in (1, 2):
        error_dict['config_yml']['biolib_version'] = \
            [f'Biolib version can only be either 1 or 2 for now, your version is {biolib_version}.']
        print_errors(error_dict)
        sys.exit(1)

    return biolib_version


def validate_modules(yaml_data, biolib_yaml_version):
    error_dict = {'modules': {}}
    if 'modules' in yaml_data:
        for name, task_data in yaml_data['modules'].items():
            task_errors = validate_module(name, task_data, biolib_yaml_version)
            if task_errors:
                error_dict['modules'].update(task_errors)

    # Just return empty dict if we have no errors
    if error_dict['modules']:
        return error_dict
    else:
        return {}


def validate_yaml_config(yaml_data, source_files_zip_path):
    source_files_zip = ZipFile(source_files_zip_path)
    error_dict = {'config_yml': {}}

    biolib_yaml_version = validate_and_get_biolib_yaml_version(yaml_data, error_dict)
    error_dict['config_yml'].update(validate_app_version(yaml_data, biolib_yaml_version, source_files_zip))
    error_dict['config_yml'].update(validate_modules(yaml_data, biolib_yaml_version))
    error_dict['config_yml'].update(validate_arguments(yaml_data))

    if error_dict['config_yml']:
        print_errors(error_dict)
        raise BiolibValidationError('Could not validate .biolib/config.yml')


def print_errors(error_dict):
    print('\nThe following validation errors were found in the .biolib/config.yml file:')
    print(yaml.safe_dump(error_dict, allow_unicode=True, default_flow_style=False))


def write_output_files(mapped_output_files=None):
    output_dir_path = 'biolib_results'

    if os.path.exists(output_dir_path):
        os.rename(output_dir_path, f'{output_dir_path}_old_{time.strftime("%Y%m%d%H%M%S")}')

    if mapped_output_files:
        logger.debug("Output Files:")

        for path, data in mapped_output_files.items():
            dirs, file = os.path.split(path)
            path_on_disk = f'{output_dir_path}{path}'

            if dirs:
                os.makedirs(f'{output_dir_path}{dirs}', exist_ok=True)
            if file:
                open(path_on_disk, 'wb').write(data)

            logger.debug(f"  - {path_on_disk}")

        logger.debug('\n')


def get_pretty_print_module_output_string(stdout=None, stderr=None, exitcode=None) -> str:
    return f'''\
stdout:
{stdout.decode()}
stderr:
{stderr.decode()}
exitcode: {exitcode}
'''


def _get_files_mappings(file_commands: List[str]) -> List[FilesMapping]:
    files_mappings: List[FilesMapping] = []
    for file_command in file_commands:
        # TODO: figure out if splitting on space is good enough
        _, from_path, to_path = file_command.split(' ')
        files_mappings.append(FilesMapping(from_path=from_path, to_path=to_path))
    return files_mappings


# Only used in run_dev which is deprecated and set to be removed
def get_mocked_app_version_from_yaml(yaml_data: Dict, source_files_zip_path: str):
    modules: List[Module] = []
    for module_name in yaml_data['modules']:
        yaml_module = yaml_data['modules'][module_name]

        image_hostname, image_uri = yaml_module['image'].split('://')
        if image_hostname != 'local-docker':
            raise Exception(
                f'Found image {image_uri} with hostname {image_hostname}, please use "local-docker://"'
                f'Your application must use modules with local docker images to use run-dev'
            )

        modules.append(Module(
            command=yaml_module.get('command', ''),
            environment='biolib-ecr',
            image_uri=image_uri,
            input_files_mappings=_get_files_mappings(yaml_module['input_files']),
            large_file_systems=[],
            name=module_name,
            output_files_mappings=_get_files_mappings(yaml_module['output_files']),
            source_files_mappings=_get_files_mappings(yaml_module.get('source_files', [])),
            working_directory=yaml_module['working_directory'],
            estimated_image_size_bytes=None  # This should not be provided here
        ))

    return AppVersionOnJob(
        client_side_executable_zip=source_files_zip_path,
        consumes_stdin=True,
        is_runnable_by_user=True,
        modules=modules,
        public_id='app-version-mock-id',
        remote_hosts=[],
        settings=[],
        stdout_render_type='text',
        main_output_file=None,
        created_at=''
    )


def create_temporary_directory_with_source_files_zip(app_path: str) -> Tuple[tempfile.TemporaryDirectory, str]:
    original_working_directory = os.getcwd()
    temporary_directory = tempfile.TemporaryDirectory()
    os.chdir(app_path)
    try:
        zip_path_without_file_extension = os.path.join(temporary_directory.name, 'biolib-cli-tmp-source-files-zip')
        zip_path = f'{zip_path_without_file_extension}.zip'
        app_folder_name = Path(app_path).absolute().parts[-1]
        shutil.make_archive(
            base_name=zip_path_without_file_extension,
            format='zip',
            root_dir='..',
            base_dir=app_folder_name,
        )

        return temporary_directory, zip_path

    except Exception as error:
        temporary_directory.cleanup()
        raise Exception('Failed to create zip of application') from error

    finally:
        # change back to old working directory
        os.chdir(original_working_directory)



from biolib.utils.cache_state import CacheState
from biolib.typing_utils import TypedDict, Optional


# TODO: Save job keys in the user state instead of a separate state file
# UuidStr = str
# class JobStateType(TypedDict):
#     job_uuid: UuidStr
#     aes_key: Optional[str]


class UserStateType(TypedDict):
    refresh_token: Optional[str]
    # jobs: Dict[UuidStr, JobStateType]


class UserState(CacheState[UserStateType]):

    @property
    def _state_path(self) -> str:
        return f'{self._user_cache_dir}/user-state.json'

    def _get_default_state(self) -> UserStateType:
        return UserStateType(refresh_token=None)



# -*- coding: utf-8 -*-
import re
import sys
from biolib import call_cli
if __name__ == '__main__':
    sys.argv[0] = re.sub(r'(-script\.pyw|\.exe)?$', '', sys.argv[0])
    sys.exit(call_cli())  # type: ignore



import base64
import os
import sys
import time
from urllib.parse import urlparse

import requests

import biolib.api

from biolib import utils
from biolib.biolib_api_client.auth import BearerAuth
from biolib.biolib_api_client import BiolibApiClient, CreatedJobDict, CloudJob
from biolib.biolib_binary_format.stdout_and_stderr import StdoutAndStderr
from biolib.biolib_errors import BioLibError, RetryLimitException, StorageDownloadFailed
from biolib.compute_node.utils import SystemExceptionCodeMap
from biolib.biolib_logging import logger
from biolib.utils import BIOLIB_PACKAGE_VERSION, IS_RUNNING_IN_NOTEBOOK
from biolib.typing_utils import List, TypedDict, Optional, Literal


class PresignedS3UploadLinkResponse(TypedDict):
    presigned_upload_url: str


class PresignedS3DownloadLinkResponse(TypedDict):
    presigned_download_url: str


def _get_user_info() -> Optional[str]:
    if utils.BASE_URL_IS_PUBLIC_BIOLIB:
        return None

    enterprise_agent_info_opt_env_vars = ['DOMINO_STARTING_USERNAME', 'USER']

    for env_var in enterprise_agent_info_opt_env_vars:
        env_var_value = os.getenv(env_var)
        if env_var_value:
            return env_var_value

    return None


class BiolibJobApi:

    @staticmethod
    def create(
            app_version_id,
            override_command=False,
            caller_job=None,
            machine='',
            experiment_uuid: Optional[str] = None
    ):
        data = {
            'app_version_id': app_version_id,
            'client_type': 'biolib-python',
            'client_version': BIOLIB_PACKAGE_VERSION,
            'client_opt_user_info': _get_user_info(),
        }

        if override_command:
            data.update({
                'arguments_override_command': override_command
            })

        if caller_job:
            data['caller_job'] = caller_job

        if machine:
            data.update({
                'requested_machine': machine
            })

        if experiment_uuid:
            data['experiment_uuid'] = experiment_uuid

        response = biolib.api._client.post(path='/jobs/', data=data)  # pylint: disable=protected-access

        return response.json()

    @staticmethod
    def update_state(job_id, state):
        response = requests.patch(
            f'{BiolibApiClient.get().base_url}/api/jobs/{job_id}/',
            json={'state': state},
            auth=BearerAuth(BiolibApiClient.get().access_token)
        )

        # TODO: Error handling with response object
        if not response.ok:
            raise BioLibError(response.content)

        return response.json()

    @staticmethod
    def create_cloud_job(job_id: str, result_name_prefix: Optional[str]) -> CloudJob:
        response = None
        data = {'job_id': job_id}
        if result_name_prefix:
            data['result_name_prefix'] = result_name_prefix

        for retry in range(4):
            try:
                response = requests.post(
                    f'{BiolibApiClient.get().base_url}/api/jobs/cloud/',
                    json=data,
                    auth=BearerAuth(BiolibApiClient.get().access_token)
                )

                if response.status_code == 503:
                    raise RetryLimitException(response.content)
                # Handle possible validation errors from backend
                elif not response.ok:
                    raise BioLibError(response.text)

                break

            except RetryLimitException as retry_exception:  # pylint: disable=broad-except
                if retry > 3:
                    raise BioLibError('Reached retry limit for cloud job creation') from retry_exception
                time.sleep(1)

        if not response:
            raise BioLibError('Could not create new cloud job')

        cloud_job: CloudJob = response.json()
        return cloud_job

    @staticmethod
    def save_compute_node_job(job, module_name, access_token, node_url):
        response = requests.post(
            f'{node_url}/v1/job/',
            json={'module_name': module_name, 'job': job, 'access_token': access_token}
        )

        if not response.ok:
            raise BioLibError(response.content)

        return response.json

    @staticmethod
    def start_local_job(job_id, module_input_serialized, node_url):
        response = requests.post(
            f'{node_url}/v1/job/{job_id}/start/',
            data=module_input_serialized,
        )

        if not response.ok:
            raise BioLibError(response.content)

    # TODO: Deprecate this
    @staticmethod
    def await_local_compute_node_status(
            retry_interval_seconds: float,
            retry_limit_minutes: float,
            status_to_await: str,
            compute_type: str,
            node_url: str,
            job: CreatedJobDict,
    ):
        status_max_retry_attempts = int(retry_limit_minutes * 60 / retry_interval_seconds)
        status_is_reached = False
        final_status_messages: List[str] = []

        def print_logs_packages(stdout_and_stderr_packages_b64):
            for stdout_and_stderr_package_b64 in stdout_and_stderr_packages_b64:
                stdout_and_stderr_package = base64.b64decode(stdout_and_stderr_package_b64)
                stdout_and_stderr = StdoutAndStderr(stdout_and_stderr_package).deserialize()

                sys.stdout.write(stdout_and_stderr.decode())
                if not IS_RUNNING_IN_NOTEBOOK:  # for some reason flushing in jupyter notebooks breaks \r handling
                    sys.stdout.flush()
            # flush after having processed all packages
            sys.stdout.flush()

        for _ in range(status_max_retry_attempts):
            response = requests.get(f'{node_url}/v1/job/{job["public_id"]}/status/')
            if not response.ok:
                raise Exception(response.content)

            status_json = response.json()

            for status_update in status_json['status_updates']:
                if status_update.get('log_message') == status_to_await:
                    status_is_reached = True

                # If the status is reached print the log messages after all stdout and stderr has been written
                if status_is_reached:
                    final_status_messages.append(status_update['log_message'])
                else:
                    # Print the status before writing stdout and stderr
                    logger.info(f'{compute_type}: {status_update["log_message"]}')

            app_version = job['app_version']
            if 'stdout_and_stderr_packages_b64' in status_json and utils.STREAM_STDOUT and \
                    (app_version.get('main_output_file') or app_version.get('stdout_render_type') == 'text'):
                print_logs_packages(status_json['stdout_and_stderr_packages_b64'])

            if 'error_code' in status_json:
                error_code = status_json['error_code']
                error_message = SystemExceptionCodeMap.get(error_code, f'Unknown error code {error_code}')

                raise BioLibError(f'{compute_type}: {error_message}')

            if status_is_reached:
                # Print the final log message after stdout and stderr has been written
                for message in final_status_messages:
                    logger.info(f'{compute_type}: {message}')

                return

            time.sleep(retry_interval_seconds)

        raise BioLibError(f'{compute_type}: Failed to get results: Retry limit exceeded')

    @staticmethod
    def get_cloud_result(job_id: str, node_url: str) -> bytes:
        for _ in range(5):
            try:
                response = requests.get(f'{node_url}/v1/job/{job_id}/result/', timeout=600)
                if response.ok:
                    return response.content
                else:
                    logger.error(
                        f'Getting result for job {job_id} failed with status {response.status_code} '
                        f'and error: {response.content.decode()}\nRetrying...'
                    )
            except Exception as error:  # pylint: disable=broad-except
                logger.error(f'Getting result for job {job_id} failed with error {error}\nRetrying...')

            time.sleep(2)

        raise BioLibError(f'Max retries hit, when getting result for job {job_id}')

    @staticmethod
    def get_enclave_json(biolib_base_url):
        response = requests.get(
            f'{biolib_base_url}/info-files/biolib-enclave.json',
        )
        return response.json()

    @staticmethod
    def get_job_storage_download_url(
            job_uuid: str,
            job_auth_token: str,
            storage_type: Literal['input', 'results'],
    ) -> str:
        response = requests.get(
            f'{BiolibApiClient.get().base_url}/api/jobs/{job_uuid}/storage/{storage_type}/download/',
            auth=BearerAuth(BiolibApiClient.get().access_token),
            headers={
                'Job-Auth-Token': job_auth_token
            }
        )

        if not response.ok:
            if response.status_code == 404:
                raise StorageDownloadFailed(response.content)
            else:
                raise BioLibError(response.content)

        presigned_s3_download_link_response: PresignedS3DownloadLinkResponse = response.json()
        presigned_download_url = presigned_s3_download_link_response['presigned_download_url']

        app_caller_proxy_job_storage_base_url = os.getenv('BIOLIB_CLOUD_JOB_STORAGE_BASE_URL', '')
        if app_caller_proxy_job_storage_base_url:
            # Done to hit App Caller Proxy when downloading result from inside an app
            parsed_url = urlparse(presigned_download_url)
            presigned_download_url = f'{app_caller_proxy_job_storage_base_url}{parsed_url.path}?{parsed_url.query}'

        return presigned_download_url
